subroutine NLCGVO(CGparam,outiter,time,nfev,ngev,inform)

	use globals  
	use myproblem

	implicit none
	
	! SCALARS ARGUMENTS
	integer,intent(in) :: CGparam
	integer,intent(out) :: outiter,nfev,ngev,inform
	real(kind=4),intent(out) :: time
	
	! LOCAL SCALARS
	integer :: ind,informLS,informALG,mquad,mnquad,&
						 nfevLS,ngevLS,allocerr,maxoutiter
	real(kind=8), parameter :: zero = 0.0d0, one = 1.0d0, two = 2.0d0
	real(kind=8) :: stp,stpmin,stpmax,epsopt,ftol,gtol,    &
					ftolLS,gtolLS,ctol,beta,fxv,fxd,fxprevvprev,       &
					fxprevv,fxdprev,fxprevdprev,norm2v,Fsupn,f,g,theta
	real(kind=4) :: timeini, timefin
	logical :: iprint,dsdc,restart

	! LOCAL ARRAYS
	real(kind=8),allocatable :: phi0(:),gphi0(:),v(:),xprev(:),vprev(:),dprev(:),JFprev(:,:)
	character(len=4) :: CGname
	character(len=1) :: restinf,restinfprev

	! EXTERNAL SUBROUTINES
	external :: evalphi,evalgphi
	
  ! This subroutine solves the multiobjective optimization problem
  !
  ! Minimize F(x)
  !
  ! where F:R^n->R^m, using the non-linear conjugate gradient method
  ! described in
	!
  ! L. R. Lucambio Pérez and L. F. Prudente, Nonlinear Conjugate 
  ! Gradient Methods for Vector Optimization, SIAM Journal on 
  ! Optimization 28(3), pp. 2690-2720, 2018.
  !
  ! The method used is chosen according to CGparam:
  !	CGparam = 0: Steepest descent method
  !	CGparam = 1: Fletcher-Reeves method
  !	CGparam = 2: Conjugate Descent method
  !	CGparam = 3: Dai-Yuan method
  !	CGparam = 4: modified Dai-Yuan method
  !	CGparam = 5: Polak-Ribière-Polyak method
  !	CGparam = 6: Hestenes-Stiefel method
  ! 
  ! The user is required to code the function values and their gradients
  ! in subroutine myproblem.
  !
  ! The steepest descent direction is calculated in innersolver routine.
  ! This routine uses the software Algencan contained in libalgencan.a.
  ! A step-size satifying the strong Wolfe conditions is determined at 
  ! each iteration by means of the line search routine lsvecopt. This
  ! routine uses the codes dcsrch and dcstep of Moré and Thuente.  
  !
  ! inform:
  !
  !  0: Optimality satisfied
  !  1: Maximum number of iterations reached
  ! -1: An error occurred during the execution of NLCG algorithm
  !
  ! November of 2017.
  !

	!---------------------------------------------------------------------   
	!     Initialization
	!---------------------------------------------------------------------
	
	! Start timing
	
	call cpu_time(timeini)	
	
	! Set default parameters 

	epsopt 	= 5.0d0 * sqrt( 2.0d0 ** (-52) )
	ftol 		= 1.d-04
	gtol		= 1.d-01
	ctol 		= 1.d-01
	
	maxoutiter = 10000

	stpmin      = 1.0d-15
	stpmax      = 1.0d+10
	
	! Allocating the arrays
	
	allocate(d(n),v(n),xprev(n),dprev(n),vprev(n),phi0(m),gphi0(m),sF(m),JF(m,n),JFprev(m,n),stat=allocerr)
	if ( allocerr .ne. 0 ) then
     write(*,*) 'Allocation error in main program'
     stop
	end if
	
	! Print problem information	
	
	iprint = .true.	
	
	if ( iprint ) then
	
			if ( CGparam == 0 ) then
				CGname = 'SD'
			else if ( CGparam == 1 ) then
				CGname = 'FR'
			else if ( CGparam == 2 ) then
				CGname = 'CD'
			else if ( CGparam == 3 ) then
				CGname = 'DY'
			else if ( CGparam == 4 ) then
				CGname = 'mDY'
			else if ( CGparam == 5 ) then
				CGname = 'PRP+'
			else if ( CGparam == 6 ) then
				CGname = 'HS+'
			end if 
	
			write(*,1000)
			write(*,1010) n, m
			write(*,1020) CGname
	end if
	
	! Check derivatives
	
	if ( checkder ) call checkd(n,m,x)
	
	! Scale problem
	
	call scalefactor(n,m,x,scaleF)	

	! Counters
	
	outiter = 0	
	nfev    = 0
	ngev    = 0
	
	! Compute v(x)
	
	do ind = 1,m
			call sevalg(n,x,JF(ind,:),ind)
			ngev = ngev + 1
	end do
	
	call evalv(n,m,v,informALG)
	
	! Compute sup-norm of v(x)
	
	norm2v = norm2(v)
	
	! Compute f(x,v)
	
	fxv = maxval( matmul( JF,v ) )
	
	! Compute theta
	
	theta = fxv + norm2v ** 2 / two
	
	if ( iprint ) then
			write(*,1030) epsopt
			write(*,1040)
			write(*,1050) outiter, abs(theta), norm2v, informALG, nfev, ngev
	end if

	
	!---------------------------------------------------------------------  
	!     Main loop
	!---------------------------------------------------------------------     

	Main_loop: do
	
		!-------------------------------------------------------------------
		!     Test stopping criteria
		!-------------------------------------------------------------------
		
		! Test optimality		
		
		if ( abs( theta ) <= epsopt  ) then
			inform = 0
			
			! Stop timing
			
			call cpu_time(timefin)
			
			time = timefin - timeini
  
			if ( iprint ) then
					if ( outiter > 0 ) then
							if ( mod(outiter,10) .eq. 0 ) write(*,1040)
							write(*,1060) outiter, abs(theta), norm2v, informLS, informALG, restinf, nfev, ngev
					end if
			
					write(*,1070) nfev, ngev, time
			end if
			
			deallocate(d,v,xprev,dprev,vprev,phi0,gphi0,sF,JF,JFprev)
			return
		end if
		
		! Test whether the number of iterations is exhausted
		
		if ( outiter >= maxoutiter ) then
				inform = 1
				
				! Stop timing
			
				call cpu_time(timefin)
			
				time = timefin - timeini
					
				if ( iprint ) then
						if ( outiter > 0 ) then
								if ( mod(outiter,10) .eq. 0 ) write(*,1040)
								write(*,1060) outiter, abs(theta), norm2v, informLS, informALG, restinf, nfev, ngev
						end if
						write(*,1080) nfev, ngev, time
				end if
				
				deallocate(d,v,xprev,dprev,vprev,phi0,gphi0,sF,JF,JFprev)
				return
		end if
		
		!-------------------------------------------------------------------
		!			Prepare the iteration
		!-------------------------------------------------------------------
		
		! Check for restart
		
		restart = .false.

		if ( outiter > 0 .and. ( informLS /= 0 .or. informALG /= 0 ) ) restart = .true.
		
		! Update parameter beta
		
		if ( outiter > 0 .and. .not. restart ) then	
		
				if ( CGparam == 0 ) then 
					! Steepest descent method:
					
					beta = zero
					
				elseif ( CGparam == 1 ) then
					! FR method:
					
					! Compute f(x,v)			
					fxv = maxval( matmul( JF,v ) ) 
					
					! Compute f(xprev,vprev)
					fxprevvprev = maxval( matmul( JFprev,vprev ) ) 
					
					beta = fxv / fxprevvprev
					beta = 0.98d0 * beta
					
				elseif ( CGparam == 2 ) then
					! CD method:
					
					! Compute f(x,v)
					fxv = maxval( matmul( JF,v ) ) 

					! Compute f(xprev,dprev)
					fxprevdprev = maxval( matmul( JFprev,dprev ) ) 
										
					beta = fxv / fxprevdprev 					
					beta = 0.99d0 * ( 1.0d0 - gtol ) * beta
					
				elseif ( CGparam == 3 ) then
					! DY method:
					
					! Compute f(x,v)
					fxv = maxval( matmul( JF,v ) )
					
					! Compute f(x,dprev)	
					fxdprev =  maxval( matmul( JF,dprev ) ) 
					
					! Compute f(xprev,dprev)
					fxprevdprev =  maxval( matmul( JFprev,dprev ) ) 
										
					beta = - fxv / ( fxdprev - fxprevdprev )
					beta = 0.99d0 * ( ( one - gtol) / ( one + gtol ) ) * beta
					
				elseif ( CGparam == 4 ) then
					! modified DY method:
					
					! Compute f(x,v)
					fxv = maxval( matmul( JF,v ) )
					
					! Compute f(x,dprev)	
					fxdprev =  maxval( matmul( JF,dprev ) ) 
					
					! Compute f(xprev,dprev)
					fxprevdprev =  maxval( matmul( JFprev,dprev ) ) 
										
					beta = - fxv / ( fxdprev - 1.02d0 * fxprevdprev )
					
				elseif ( CGparam == 5 ) then
					! PRP+ method:
					
					! Compute f(xprev,vprev)
					fxprevvprev = maxval( matmul( JFprev,vprev ) ) 
					
					! Compute f(x,v)
					fxv = maxval( matmul( JF,v ) )
					
					! Compute f(xprev,v)
					fxprevv = maxval( matmul( JFprev,v ) )
					
					beta = ( - fxv + fxprevv ) / ( - fxprevvprev )
					beta = max( zero, beta )
					
				elseif ( CGparam == 6 ) then
					! HS+ method:	
					
					! Compute f(x,v)
					fxv = maxval( matmul( JF,v ) )
					
					! Compute f(xprev,v)
					fxprevv = maxval( matmul( JFprev,v ) )
					
					! Compute f(x,dprev)
					fxdprev =  maxval( matmul( JF,dprev ) ) 
					
					! Compute f(xprev,dprev)
					fxprevdprev = maxval( matmul( JFprev,dprev ) ) 

					beta = ( - fxv + fxprevv ) / ( fxdprev - fxprevdprev )
					beta = max( zero, beta )
				end if
				
		end if
		
		! Define the search direction
		
		if ( outiter == 0 .or. restart ) then
				d = v
		else
				d = v + beta * d
		end if
		
		! Identify a steepest descent iteration
		
		if ( outiter > 0 ) restinfprev = restinf
		
		if ( outiter == 0 .or. beta == 0.0d0 .or. restart ) then 
				restinf = 'Y'
		else 
				restinf = 'N'
		end if
		
		! Test if direction d satisfies the sufficient descent condition
		
		dsdc = .true.
		
		if ( outiter > 0 .and. 										   &
							 ( CGparam == 1 .or. CGparam == 5 .or. CGparam == 6 ) ) then
		
				fxd = maxval( matmul( JF,d ) )
				fxv = maxval( matmul( JF,v ) )
				
				if ( fxd > ctol * fxv ) then
						dsdc = .false.
				end if
		end if

		!-------------------------------------------------------------------
		!     Iteration
		!-------------------------------------------------------------------
		
		if ( dsdc ) then

				! Print information
		
				if ( iprint .and. outiter > 0 ) then
						if ( mod(outiter,10) .eq. 0 ) write(*,1040)
						write(*,1060) outiter, abs(theta), norm2v, informLS, informALG, restinfprev, nfev, ngev
				end if
				
				! Increment outiter
				
				outiter = outiter + 1
						
				! Compute initial step
				
				if ( outiter == 1 ) then
						stp = one/norm2v
						stp = max( stp , one    )
						stp = min( stp , stpmax )
				else
					
						fxd = maxval( matmul( JF,d ) )
						fxprevdprev = maxval( matmul( JFprev,dprev ) )
						
						stp = stp * fxprevdprev/fxd
						stp = max( stp , 1.0d-2 )
						stp = min( stp , stpmax, 1.0d2 )
				end if
				
				! Define the tolerances for the line search algorithm

				ftolLS = ftol
				gtolLS = gtol
		
				! Compute phi0 = [phi_1(x),...,phi_m(x)] 
				
				do ind = 1, m
					call sevalf(n,x,phi0(ind),ind)
					nfev = nfev + 1   
				end do
				
				! Compute gphi0 = [gphi_1(x),...,gphi_m(x)]
				
				gphi0 = matmul( JF, d )		
					
		else
				! Retrieve previous data
				
				x  = xprev
				v  = vprev
				d  = dprev
				JF = JFprev
				
				! Define the tolerances for the line search algorithm

				gtolLS = 1.0d-1 * gtolLS
				
				if ( gtolLS <= 5.0d0 * ftolLS ) ftolLS = 1.0d-1 * ftolLS
				
				if ( ftolLS == 1.0d-16 .and. informLS /= 0 ) then
						return
				end if
		
		end if		
		
		! Compute the stepsize satisfying the strong Wolfe conditions
		
		call lsvecopt(evalphi,evalgphi,stp,stpmin,stpmax,m,phi0,gphi0,     &
		              ftolLS,gtolLS,nfevLS,ngevLS,informLS)            
		
		! Check for errors in lsvecopt
		
		if ( informLS == - 1 ) then
				inform = - 1
				
				deallocate(d,v,xprev,dprev,vprev,phi0,gphi0,sF,JF,JFprev)
				return
		end if
				
							  
		nfev = nfev + nfevLS
		ngev = ngev + ngevLS
		
		! Save previous data				
		
		xprev  = x
		vprev  = v
		dprev  = d
		JFprev = JF
		
		! Update x	
		
		x = x + stp * d
				
		! Compute v(x)
		
		do ind = 1,m
			call sevalg(n,x,JF(ind,:),ind)
			ngev = ngev + 1
		end do
		
		call evalv(n,m,v,informALG)
		
		! Compute sup-norm of v(x)
	
		norm2v = norm2(v)
		
		! Compute f(x,v)
	
		fxv = maxval( matmul( JF,v ) )
		
		! Compute theta
		
		theta = fxv + norm2v ** 2 / two
		
	!---------------------------------------------------------------------
	!     Iterate
	!---------------------------------------------------------------------    

	end do Main_loop

	!--------------------------------------------------------------------- 
	!     End of main loop
	!---------------------------------------------------------------------
	
	! Non-executable statements
	
	1000 format(/,1X,'------------------------------------------------------------',/,1X,& 
					 'NLCG: Non-linear conjugate grandient for vector optimization',/,1X,& 
					 '------------------------------------------------------------')
									 
	1010 format(/,1X,'Number of variables:',1X,I5,/,1X,'Number of functions:',1X,I5)	

	1020 format(/,1X,'CG parameter:',1X,4A)	
	
	1030 format(/,1X,'Optimality tolerance:',1X,1P,D7.1)
	
	1040 format(/,4X,'out',2X,'|theta|',4X,'norm(v)',3X,'LS',1X,'ALG',1X,'SD',1X,'#evalf',2X,'#evalg')
	
	1050 format(1X,I5,3X,1P,D8.2,3X,1P,D8.2,3X,'-',2X,I1,3X,'-',1X,I6,2X,I6)
	
	1060 format(1X,I5,3X,1P,D8.2,3X,1P,D8.2,3X,I1,2X,I1,3X,1A,1X,I6,2X,I6)		
	
	1070 format(/,1X,'Flag of NLCG: solution was found',/,/,1X, &
							'Number of functions evaluations:   ',I6,/,1X, &
							'Number of derivatives evaluations: ',I6/,/,1X, &
							'Total CPU time in seconds: ',0P,F8.2)						
	
	1080 format(/,1X,'Flag of NLCG: maximum of iterations reached',/,/,1X,&
							'Number of functions evaluations:   ',I6,/,1X, &
							'Number of derivatives evaluations: ',I6/,/,1X, &
							'Total CPU time in seconds: ',0P,F8.2)

end subroutine NLCGVO
	
!***********************************************************************
!***********************************************************************

subroutine scalefactor(n,m,x,scaleF)	

	use globals, only: sF
	use myproblem
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in)  :: n,m
	logical, intent(in)  :: scaleF

	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n)
		
	! LOCAL SCALAR
	integer :: i,ind
	real(kind=8) :: eps
	
	! LOCAL ARRAY
	real(kind=8) :: g(n)
	
	eps = 1.0d-08

! Compute objective functions scaling factor

	if ( scaleF ) then
			do ind = 1,m
					call evalg(n,x,g,ind)

					sF(ind) = 1.0d0
					do i = 1,n
							sF(ind) = max( sF(ind), abs( g(i) ) )
					end do
					
					sF(ind) = max( eps, 1.0d0 / sF(ind) )
			end do
	else
			sF(1:m) = 1.0d0
	end if
	
end subroutine scalefactor

!***********************************************************************
!***********************************************************************

subroutine sevalf(n,x,f,ind)

	use globals, only: sF
	use myproblem

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in) :: n,ind
	real(kind=8), intent(in) :: x(n)
	real(kind=8), intent(out) :: f
	
	call evalf(n,x,f,ind)
	
	f = sF(ind) * f
	
end subroutine sevalf

!***********************************************************************
!***********************************************************************

subroutine sevalg(n,x,g,ind)

	use globals, only: sF
	use myproblem
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in) :: n,ind
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n)
	real(kind=8), intent(out) :: g(n)
	
	call evalg(n,x,g,ind)
	
	g = sF(ind) * g
	
end subroutine sevalg


!***********************************************************************
!***********************************************************************

subroutine evalphi(stp,phi,ind)

	use globals, only: n,x,d

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in) :: ind
	real(kind=8), intent(in) :: stp
	real(kind=8), intent(out) :: phi

	call sevalf(n,x + stp * d,phi,ind)

end subroutine evalphi

!***********************************************************************
!***********************************************************************

subroutine evalgphi(stp,gphi,ind)

	use globals, only: n,x,d

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in) :: ind
	real(kind=8), intent(in) :: stp
	real(kind=8), intent(out) :: gphi

	! LOCAL ARRAYS
	real(kind=8) :: g(n)

	call sevalg(n,x + stp * d,g,ind)

	gphi = dot_product(g,d)

end subroutine evalgphi
